#include <stdio.h>
#include <string.h>
#include <time.h>

#include "parser.h"

#include "consola.h"
#include "symtab.h"

extern long *lpErrors;
extern long lnErrors;
extern bool bError;
extern long lSymbolIndex[7];
extern _symtab symtab[7];

extern bool bLog;

extern time_t	tTime;

long reporterrors(char *psz)
	{
	FILE *fp;
	char szErrorText[0x100];
	long lError;
	long l;
	_symtab *ps = &symtab[6];
	bool bFound;

	if (bLog)
		{
		fp = fopen("errors.txt", "a");
		if (!fp)
			{
			bLog = false;
			printf("Error opening log file \"errors.txt\"");
			goto NoLog;
			}
		fprintf(fp, "Errors for %s, %s\n-==--==--==--==--==--==--==--==--==--==--==--==--==--==--==--==-\n", psz, ctime(&tTime));
		}
NoLog:

	for (long i = 0; i < lnErrors; i++)
		{
		lError = lpErrors[i] & 255;
		switch (lError)
			{
			case EXPECTED_STATIC_VAR_DECL:
				strcpy(szErrorText, "Expected static-var declaration");
				break;
			case EXPECTED_PIECE_DECL:
				strcpy(szErrorText, "Expected piece declaration");
				break;
			case EXPECTED_PROC_DECL:
				strcpy(szErrorText, "Expected proc declaration");
				break;
			case EXPECTED_IDENTIFIER_LIST:
				strcpy(szErrorText, "Expected identifier list");
				break;
			case EXPECTED_END_STMT:
				strcpy(szErrorText, "Expected ';'");
				break;
			case SPECIFIED_NONE_AFTER_DECL:
				strcpy(szErrorText, "Specified \"none\" after declaration");
				break;
			case EXPECTED_IDENTIFIER:
				strcpy(szErrorText, "Expected identifier");
				break;
			case EXPECTED_PROC_NAME:
				strcpy(szErrorText, "Expected proc name");
				break;
			case DIDNT_SPECIFY_ARGS:
				strcpy(szErrorText, "Didn't specify args.  Use \"args none;\" if no arguments");
				break;
			case EXPECTED_ARGLIST:
				strcpy(szErrorText, "Expected argument list");
				break;
			case DIDNT_SPECIFY_VAR:
				strcpy(szErrorText, "Didn't specify var.  Use \"var none;\" if no var");
				break;
			case EXPECTED_VARLIST:
				strcpy(szErrorText, "Expected var list");
				break;
			case EXPECTED_BEGINPROC:
				strcpy(szErrorText, "Expected \"beginproc\"");
				break;
			case EXPECTED_PROC_START:
				strcpy(szErrorText, "Expected proc start");
				break;
			case EXPECTED_PROC_BODY:
				strcpy(szErrorText, "Expected proc body");
				break;
			case EXPECTED_PROC_END:
				strcpy(szErrorText, "Expected proc end");
				break;
			case EXPECTED_ENDPROC:
				strcpy(szErrorText, "Expected \"endproc;\"");
				break;
			case END_STMT_AFTER_BEGINPROC:
				strcpy(szErrorText, "Found ';' after \"beginproc\"");
				break;
			case MISSING_END_STMT_AFTER_ENDPROC:
				strcpy(szErrorText, "Missing ';' after \"endproc\"");
				break;
			case TOO_MANY_CONVERSIONS:
				strcpy(szErrorText, "Too many conversions");
				break;
			case CANT_CONVERT_TWICE:
				strcpy(szErrorText, "Can't convert twice");
				break;
			case INCOMPATIBLE_CONVERSION_TYPES:
				strcpy(szErrorText, "Incompatible conversion types");
				break;
			case DUPLICATE_EXPORT_SYMBOL:
				strcpy(szErrorText, "Duplicate DLL Export Referenced (Symbol Collision)");
				break;
			case DUPLICATE_DLL_NAME:
				strcpy(szErrorText, "Duplicate DLL Name Referenced (Symbol Collision)");
				break;
			case EXPECTED_OPEN_BRACE:
				strcpy(szErrorText, "Expected: \'{\'");
				break;
			case EXPECTED_DLL_EXPORT_LIST:
				strcpy(szErrorText, "Expected DLL Export List");
				break;
			case EXPECTED_DLL_NAME:
				strcpy(szErrorText, "Expected DLL Name");
				break;
			case EXPECTED_KW_DLL:
				strcpy(szErrorText, "Expected \"dll\"");
				break;
			case EXPECTED_LINKED_FUNCTION:
				strcpy(szErrorText, "Expected Linked Function");
				break;
			case INVALID_DLLLINK_SYMBOL:
				sprintf(szErrorText, "Invalid DLL Link Symbol, %s", ps->tok.psz);
				break;
			case INVALID_STRING:
				sprintf(szErrorText, "Invalid String %s", ps->tok.psz);
				break;
			case ERROR_AT_TOP_LEVEL:
				strcpy(szErrorText, "Error At Top Level");
				break;
			case UNRECOGNIZED_TERM:
				{
				l = (lpErrors[i] >> 8) & 255;
				bFound = false;
				for (long j = 0; (j < lSymbolIndex[6]) && !bFound; j++)
					{
					if (ps->lIndex == l) bFound = true;
					else ps = ps->pNext;
					}
				sprintf(szErrorText, "Unrecognized term, %s", ps->tok.psz);
				}
				break;
			case UNDEFINED_TERM:
				{
				l = (lpErrors[i] >> 8) & 255;
				bFound = false;
				for (long j = 0; (j < lSymbolIndex[6]) && !bFound; j++)
					{
					if (ps->lIndex == l) bFound = true;
					else ps = ps->pNext;
					}
				sprintf(szErrorText, "Undefined term, %s", ps->tok.psz);
				}
				break;
			case WARNINGS_AS_ERRORS:
				strcpy(szErrorText, "Warnings treated as errors (did you use the -e switch by mistake?)");
				break;
			}

		if (bLog)
			fprintf(fp, "Error on line %d: 0x%02X (%s)\n", lpErrors[i] >> 16, lError, szErrorText);

		textcolor(LIGHTRED);
		printf("Error on line %d: 0x%02X (%s)\n", lpErrors[i] >> 16, lError, szErrorText);
		}

	textcolor(LIGHTGRAY);
	
	if (bLog)
		{
		fprintf(fp, "\n\n\n");
		fclose(fp);
		}	

	return 0;
	}