#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include "parser.h"

#include "symtab.h"
#include "tokenizer.h"
#include "warningreporter.h"

long lDeclType;

_symtab symtab[7];
//_symtab *psymtab[6] = { &symtab[0], &symtab[1], &symtab[2], &symtab[3], &symtab[4], &symtab[5] };
long lSymbolIndex[7] = { 0, 0, 0, 0, 0, 0, 0 };

_dlllink	*LinkedDLLProcs;
long		lnDLLLinks;

_stringlink	*LinkedStrings;
long		lnStringLinks;

extern long lnPieces, lnStaticVar, lnFunctions;

bool bEOF;
bool bOK;
bool bJump;

long *lpErrors;
long lnErrors;
bool bError;
bool bDecl;
bool bInvoke;
bool bPushString;
bool bOutputStringWithoutPush;

long lnWarnings;
char **ppszWarningText;

_token *BASMFile(_token *p)
	{
	lnErrors = 0;
	bError = false;
	bEOF = false;
	bDecl = false;
	bInvoke = false;
	bPushString = false;
	bOutputStringWithoutPush = false;
	lnDLLLinks = 0;
	lnStringLinks = 0;
	lnWarnings = 0;

	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	_token *pTok = p;
	_token *pResult;

	pResult = BASMSetup(pTok);
	if (!pResult) return NULL;

	pTok = pResult;
	pResult = BASMGlobals(pTok);

	if (!pResult && !bEOF)
		{
		lnErrors++;
		lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
		lpErrors[lnErrors - 1] = ERROR_AT_TOP_LEVEL;
		lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
		return NULL;
		}

	return pResult;
	}

_token *BASMSetup(_token *p)
	{
	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	return p;
	}

_token *BASMGlobals(_token *p)
	{
	_token *pTok = p;
	_token *pResult;
	_token *pOK = NULL;

Start:
	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	pResult = StaticVarDecl(pTok);
	if (bError)
		{
		lnErrors++;
		lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
		lpErrors[lnErrors - 1] = EXPECTED_STATIC_VAR_DECL;
		lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
		return NULL;
		}
	if (!pResult) pResult = PieceDecl(pTok);
	if (bError)
		{
		lnErrors++;
		lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
		lpErrors[lnErrors - 1] = EXPECTED_PIECE_DECL;
		lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
		return NULL;
		}
	if (!pResult) pResult = ProcDecl(pTok);
	if (bError)
		{
		lnErrors++;
		lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
		lpErrors[lnErrors - 1] = EXPECTED_PROC_DECL;
		lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
		}
	if (!pResult) return pOK;

	pTok = pResult;
	pOK = pResult;
	goto Start;
	}

_token *StaticVarDecl(_token *p)
	{
	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	lDeclType = 1;

	if (strcmp(p->psz, "static-var")) return NULL;

	p->lType = KW_STATIC_VAR;

	_token *pTok = p->pNext;
	_token *pResult = IdentifierList(pTok);

	if (bError)
		{
		lnErrors++;
		lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
		lpErrors[lnErrors - 1] = EXPECTED_IDENTIFIER_LIST;
		lpErrors[lnErrors - 1] |= (p->lLine << 16);
		}

	return pResult;
	}

_token *PieceDecl(_token *p)
	{
	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	lDeclType = 0;

	if (strcmp(p->psz, "piece")) return NULL;

	p->lType = KW_PIECE;

	_token *pTok = p->pNext;
	_token *pResult = IdentifierList(pTok);

	if (bError)
		{
		lnErrors++;
		lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
		lpErrors[lnErrors - 1] = EXPECTED_IDENTIFIER_LIST;
		lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
		}

	return pResult;
	}

_token *ProcDecl(_token *p)
	{
	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	_token *pTok = p;
	_token *pResult = ProcStart(pTok);

	if (bError)
		{
		lnErrors++;
		lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
		lpErrors[lnErrors - 1] = EXPECTED_PROC_START;
		lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
		}

	if (!pResult) return NULL;

	pTok = pResult;
	pResult = ProcBody(pTok);

	if (bError || (!pResult))
		{
		lnErrors++;
		lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
		lpErrors[lnErrors - 1] = EXPECTED_PROC_BODY;
		lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
		}

	if (!pResult) return NULL;

	pTok = pResult;
	pResult = ProcEnd(pTok);

	if (!bEOF && (bError || (!pResult)))
		{
		lnErrors++;
		lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
		lpErrors[lnErrors - 1] = EXPECTED_PROC_END;
		lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
		}

	return pResult;
	}

_token *IdentifierList(_token *p)
	{
	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	_token *pTok = p;
	_token *pResult;
	long lnTimes = 0;
	
Start:
	if (!strcmp(pTok->psz, "none"))
		{
		pTok->lType = NOCODE;
		if (lnTimes)
			{
			bError = true;
			lnErrors++;
			lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
			lpErrors[lnErrors - 1] = SPECIFIED_NONE_AFTER_DECL;
			lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
			return NULL;
			}
		pTok = pTok->pNext;
		if ((strlen(pTok->psz) == 1) && pTok->psz[0] == ';')
			{
			pTok->lType = END_STMT;
			return pTok->pNext;
			}
		else
			{
			bError = true;
			lnErrors++;
			lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
			lpErrors[lnErrors - 1] = EXPECTED_END_STMT;
			lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
			return NULL;
			}
		}

	pResult = Identifier(pTok);
	if (!pResult)
		{
		bError = true;
		lnErrors++;
		lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
		lpErrors[lnErrors - 1] = EXPECTED_IDENTIFIER;
		lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
		return NULL;
		}

	lnTimes++;
	pTok = pResult;
	if ((strlen(pTok->psz) == 1) && pTok->psz[0] == ',')
		{
		pTok->lType = COMMA;
		pTok = pTok->pNext;
		goto Start;
		}

	if ((strlen(pTok->psz) == 1) && pTok->psz[0] == ';')
		{
		pTok->lType = END_STMT;
		if (lDeclType == 0) //piece decl
			lnPieces += lnTimes;
		else if (lDeclType == 1) //static-var
			lnStaticVar += lnTimes;
		return pTok->pNext;
		}
	else
		{
		bError = true;
		lnErrors++;
		lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
		lpErrors[lnErrors - 1] = EXPECTED_END_STMT;
		lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
		return NULL;
		}

//	return NULL;
	}

long lChrs;

char *pPushStrString;

_token *IsValidString(_token *p)
	{
	pPushStrString = NULL;
	_token *pTok = p;

	if ((strlen(pTok->psz) < 2) || (pTok->psz[0] != '\"') || (pTok->psz[strlen(pTok->psz) - 1] != '\"'))
		return NULL;

	pPushStrString = new char [strlen(pTok->psz) - 1];
	memcpy(pPushStrString, pTok->psz + 1, strlen(pTok->psz) - 2);
	pPushStrString[strlen(pTok->psz) - 2] = NULL;

	memcpy(pTok->psz, pTok->psz + 1, strlen(pTok->psz) - 2); //remove the wrapping quotes from the origional string
	pTok->psz[strlen(pTok->psz) - 2] = NULL;

	return pTok;
	}

_token *IsValidInvokeIdentifier(_token *p)
	{
	bool bAccess = false;
	lChrs = 0;
	_token *pTok = p;

	if (!(isalpha(pTok->psz[0]) || pTok->psz[0] == '_')) return NULL;
	lChrs = 1;

	long lLen = strlen(pTok->psz);

	for (long i = 0; i < lLen; i++)
		{
		if (!(isalnum(pTok->psz[i]) || (pTok->psz[i] == '_') || (pTok->psz[i] == '.')))
			{
			if ((pTok->psz[i] == '-') && (pTok->psz[i + 1] == '>'))
				{
				if (bAccess) //if we already found '->'
					return NULL;
				bAccess = true;
				i++;
				}
			else if (pTok->psz[i] != '-')
				return NULL;
			}
		else
			lChrs++;
		}

	return pTok;
	}

_token *IsValidIdentifier(_token *p)
	{
	_token *pTok = p;

	if (!(isalpha(pTok->psz[0]) || pTok->psz[0] == '_')) return NULL;

	long lLen = strlen(pTok->psz);

	for (long i = 0; i < lLen; i++)
		if (!(isalnum(pTok->psz[i]) || pTok->psz[i] == '_' || pTok->psz[i] == '-')) return NULL;

	return pTok;
	}

_token *Identifier(_token *p)
	{
	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	_token *pTok = p;

	if (!(isalpha(pTok->psz[0]) || pTok->psz[0] == '_')) return NULL;

	long lLen = strlen(pTok->psz);

	for (long i = 0; i < lLen; i++)
		if (!(isalnum(pTok->psz[i]) || pTok->psz[i] == '_' || pTok->psz[i] == '-')) return NULL;

	if (!strcmp(pTok->psz, "none") && bDecl)
		{
		bError = true;
		lnErrors++;
		lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
		lpErrors[lnErrors - 1] = SPECIFIED_NONE_AFTER_DECL;
		lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
		return NULL;
		}

	_symtab *ps = &symtab[lDeclType];
	bool bFound = false;
	for (long l = 0; (l < lSymbolIndex[lDeclType]) && !bFound; l++)
		{
		if (ps->tok.psz && !strcmp(ps->tok.psz, pTok->psz)) bFound = true;
		else ps = ps->pNext;
		}

	if (!bFound)
		{
		ps->pNext = (_symtab *)malloc(sizeof(_symtab));
		ps->pNext->tok.psz = NULL;
		ps->pNext->pParent = ps;
		ps->pNext->pNext = NULL;
		ps->lIndex = lSymbolIndex[lDeclType]++;
		ps->tok.psz = (char *)malloc(strlen(pTok->psz) + 1);
		strcpy(ps->tok.psz, pTok->psz);
		ps = ps->pNext;
		}

	if (lDeclType == 2) pTok->lType = NEWPROC;
	else if (lDeclType == 1) pTok->lType = DECL_STATIC_VAR;
	else if (lDeclType == 0) pTok->lType = DECL_PIECE;
	else if (lDeclType == 3) pTok->lType = CREATELOCALVAR;
	else if (lDeclType == 4) pTok->lType = CREATELOCALVAR;

	return pTok->pNext;
	}

_token *ProcStart(_token *p)
	{
	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	lDeclType = 2;

	_token *pTok = p;
	_token *pResult;
	if (strcmp(pTok->psz, "proc")) return NULL;

	p->lType = KW_PROC;

	pTok = pTok->pNext;
	pResult = Identifier(pTok);

	if (!pResult)
		{
		bError = true;
		lnErrors++;
		lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
		lpErrors[lnErrors - 1] = EXPECTED_PROC_NAME;
		lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
		return NULL;
		}

	pTok = pResult;
	if (strcmp(pTok->psz, "args"))
		{
		bError = true;
		lnErrors++;
		lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
		lpErrors[lnErrors - 1] = DIDNT_SPECIFY_ARGS;
		lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
		return NULL;
		}

	pTok->lType = KW_ARGS;

	pTok = pTok->pNext;
	if (!strcmp(pTok->psz, "none"))
		{
		pTok->lType = NOCODE;
		pResult = pTok->pNext;
		pTok = pResult;
		if (!((strlen(pTok->psz) == 1) && pTok->psz[0] == ';'))
			{
			bError = true;
			lnErrors++;
			lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
			lpErrors[lnErrors - 1] = EXPECTED_END_STMT;
			lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
			return NULL;
			}
		pTok->lType = END_STMT;
		pTok = pTok->pNext;
		}
	else
		{
		lDeclType = 3;
		pResult = ArgList(pTok);
		if (pResult)
			{
			pTok = pResult;
			if (!((strlen(pTok->psz) == 1) && pTok->psz[0] == ';'))
				{
				bError = true;
				lnErrors++;
				lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
				lpErrors[lnErrors - 1] = EXPECTED_END_STMT;
				lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
				return NULL;
				}
			pTok->lType = END_STMT;
			pTok = pTok->pNext;
			}
		else
			{
			bError = true;
			lnErrors++;
			lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
			lpErrors[lnErrors - 1] = EXPECTED_ARGLIST;
			lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
			return NULL;
			}
		}

	if (strcmp(pTok->psz , "var"))
		{
		bError = true;
		lnErrors++;
		lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
		lpErrors[lnErrors - 1] = DIDNT_SPECIFY_VAR;
		lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
		return NULL;
		}

	pTok->lType = KW_VAR;
	pTok = pTok->pNext;
	if (!strcmp(pTok->psz, "none"))
		{
		pTok->lType = NOCODE;
		pResult = pTok->pNext;
		pTok = pResult;
		if (!((strlen(pTok->psz) == 1) && pTok->psz[0] == ';'))
			{
			bError = true;
			lnErrors++;
			lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
			lpErrors[lnErrors - 1] = EXPECTED_END_STMT;
			lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
			return NULL;
			}
		pTok->lType = END_STMT;
		pTok = pTok->pNext;
		}
	else
		{
		lDeclType = 3;
		pResult = VarList(pTok);
		if (pResult)
			{
			pTok = pResult;
			if (!((strlen(pTok->psz) == 1) && pTok->psz[0] == ';'))
				{
				bError = true;
				lnErrors++;
				lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
				lpErrors[lnErrors - 1] = EXPECTED_END_STMT;
				lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
				return NULL;
				}
			pTok->lType = END_STMT;
			pTok = pTok->pNext;
			}
		else
			{
			bError = true;
			lnErrors++;
			lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
			lpErrors[lnErrors - 1] = EXPECTED_VARLIST;
			lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
			return NULL;
			}
		}

	if (strcmp(pTok->psz, "beginproc"))
		{
		bError = true;
		lnErrors++;
		lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
		lpErrors[lnErrors - 1] = EXPECTED_BEGINPROC;
		lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
		return NULL;
		}

	if ((strlen(pTok->psz) == 1) && (pTok->psz[0] == ';'))
		{
		bError = true;
		lnErrors++;
		lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
		lpErrors[lnErrors - 1] = END_STMT_AFTER_BEGINPROC;
		lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
		return NULL;
		}


	pTok->lType = NOCODE;

	return pTok->pNext;
	}

_token *ProcBody(_token *p)
	{
	_token *pTok = p;
	_token *pResult;
	_token *pOK = NULL;

Start:
	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	pResult = BASMStmt(pTok);
	if (!pResult) pResult = Label(pTok);
	if (!pResult)
		{
		if (!pOK && strcmp(pTok->psz, "endproc"))
			{
			lnErrors++;
			lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
			lpErrors[lnErrors - 1] = UNRECOGNIZED_TERM;
			_symtab *ps = &symtab[6];
			bool bFound = false;
			for (long i = 0; (i < lSymbolIndex[6]) && !bFound; i++)
				{
				if (!strcmp(ps->tok.psz, pTok->psz)) bFound = true;
				else ps = ps->pNext;
				}

			if (!bFound)
				{
				ps->tok.psz = new char[strlen(pTok->psz) + 1];
				strcpy(ps->tok.psz, pTok->psz);
				ps->pNext = (_symtab *)malloc(sizeof(_symtab));
				ps->pNext->tok.psz = NULL;
				ps->pNext->pParent = ps;
				ps->pNext->pNext = NULL;
				ps->lIndex = lSymbolIndex[6]++;
				}

			lpErrors[lnErrors - 1] |= (ps->lIndex << 8);
			lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
			}
		if (strcmp(pTok->psz, "endproc"))
			{
			lnErrors++;
			lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
			lpErrors[lnErrors - 1] = UNRECOGNIZED_TERM;
			_symtab *ps = &symtab[6];
			bool bFound = false;
			for (long i = 0; (i < lSymbolIndex[6]) && !bFound; i++)
				{
				if (!strcmp(ps->tok.psz, pTok->psz)) bFound = true;
				else ps = ps->pNext;
				}

			if (!bFound)
				{
				ps->tok.psz = new char[strlen(pTok->psz) + 1];
				strcpy(ps->tok.psz, pTok->psz);
				ps->pNext = (_symtab *)malloc(sizeof(_symtab));
				ps->pNext->tok.psz = NULL;
				ps->pNext->pParent = ps;
				ps->pNext->pNext = NULL;
				ps->lIndex = lSymbolIndex[6]++;
				}

			lpErrors[lnErrors - 1] |= (ps->lIndex << 8);
			lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
			}
		return pOK;
		}

	pTok = pResult;
	pOK = pResult;
	goto Start;
	}

_token *ProcEnd(_token *p)
	{
	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	_token *pTok = p;
	if (strcmp(pTok->psz, "endproc"))
		{
		lnErrors++;
		lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
		lpErrors[lnErrors - 1] = EXPECTED_ENDPROC;
		lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
		return NULL;
		}

	pTok->lType = NOCODE;

	pTok = p->pNext;
	if (!((strlen(pTok->psz) == 1) && pTok->psz[0] == ';'))
		{
		lnErrors++;
		lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
		lpErrors[lnErrors - 1] = MISSING_END_STMT_AFTER_ENDPROC;
		lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
		return NULL;
		}

	pTok->lType = END_STMT;

	lnFunctions++;

	if (!pTok->pNext) bEOF = true;

	return pTok->pNext;
	}

_token *BASMStmt(_token *p)
	{
	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	bJump = false;

	_token *pTok = p;
	_token *pResult;

	pResult = kwOp(pTok);
	if (!pResult) return NULL;

	pTok = pResult;
	//pResult = Args(pTok);
	pResult = Arg(pTok);

	if (!pResult) pResult = pTok;

	if (!((strlen(pResult->psz) == 1) && pResult->psz[0] == ';'))
		{
		lnErrors++;
		lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
		lpErrors[lnErrors - 1] = EXPECTED_END_STMT;
		lpErrors[lnErrors - 1] |= (pResult->lLine << 16);
		return NULL;
		}

	pResult->lType = END_STMT;

	return pResult->pNext;
	}

_token *kwOp(_token *p)
	{
	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	_token *pTok = p;

	if (!strcmp(pTok->psz, "move"))
		pTok->lType = MOVE;
	else if (!strcmp(pTok->psz, "turn"))
		pTok->lType = TURN;
	else if (!strcmp(pTok->psz, "spin"))
		pTok->lType = SPIN;
	else if (!strcmp(pTok->psz, "stop-spin"))
		pTok->lType = STOPSPIN;
	else if (!strcmp(pTok->psz, "show"))
		pTok->lType = SHOW;
	else if (!strcmp(pTok->psz, "hide"))
		pTok->lType = HIDE;
	else if (!strcmp(pTok->psz, "cache"))
		pTok->lType = CACHE;
	else if (!strcmp(pTok->psz, "dont-cache"))
		pTok->lType = DONTCACHE;
	else if (!strcmp(pTok->psz, "shade"))
		pTok->lType = SHADE;
	else if (!strcmp(pTok->psz, "dont-shade"))
		pTok->lType = DONTSHADE;
	else if (!strcmp(pTok->psz, "sfx"))
		pTok->lType = SFX;
	else if (!strcmp(pTok->psz, "wait-for-turn"))
		pTok->lType = WAITFORTURN;
	else if (!strcmp(pTok->psz, "wait-for-move"))
		pTok->lType = WAITFORMOVE;
	else if (!strcmp(pTok->psz, "sleep"))
		pTok->lType = SLEEP;
	else if (!strcmp(pTok->psz, "rand"))
		pTok->lType = RAND;
	else if (!strcmp(pTok->psz, "get"))
		pTok->lType = GET;
	else if (!strcmp(pTok->psz, "getuv"))
		pTok->lType = GETUV;
	else if (!strcmp(pTok->psz, "start"))
		pTok->lType = START;
	else if (!strcmp(pTok->psz, "call"))
		pTok->lType = CALL;
	else if (!strcmp(pTok->psz, "ret"))
		pTok->lType = RET;
	else if (!strcmp(pTok->psz, "sig"))
		pTok->lType = SIG;
	else if (!strcmp(pTok->psz, "signal"))
		pTok->lType = SIG;
	else if (!strcmp(pTok->psz, "mask"))
		pTok->lType = MASK;
	else if (!strcmp(pTok->psz, "explode"))
		pTok->lType = EXPLODE;
	else if (!strcmp(pTok->psz, "set"))
		pTok->lType = SET;
	else if (!strcmp(pTok->psz, "attach"))
		pTok->lType = ATTACH;
	else if (!strcmp(pTok->psz, "drop"))
		pTok->lType = DROP;
	else if (!strcmp(pTok->psz, "nop"))
		pTok->lType = NOP;
	else if (!strcmp(pTok->psz, "empty"))
		pTok->lType = EMPTY;
	else if (!strcmp(pTok->psz, "ddw"))
		pTok->lType = DDW;
	else if (!strcmp(pTok->psz, "pushc"))
		pTok->lType = PUSHCONSTANT;
	else if (!strcmp(pTok->psz, "pushl"))
		pTok->lType = PUSHVAR;
	else if (!strcmp(pTok->psz, "pushs"))
		pTok->lType = PUSHSTATIC;
	else if (!strcmp(pTok->psz, "popl"))
		pTok->lType = POPVAR;
	else if (!strcmp(pTok->psz, "pops"))
		pTok->lType = POPSTATIC;
	else if (!strcmp(pTok->psz, "clv"))
		pTok->lType = CREATELOCALVAR;
	else if (!strcmp(pTok->psz, "flush"))
		pTok->lType = STACKFLUSH;
	else if (!strcmp(pTok->psz, "jmp"))
		{
		bJump = true;
		pTok->lType = JMP;
		}
	else if (!strcmp(pTok->psz, "je"))
		{
		bJump = true;
		pTok->lType = JMP;
		}
	else if (!strcmp(pTok->psz, "jz"))
		{
		bJump = true;
		pTok->lType = JMP;
		}
	else if (!strcmp(pTok->psz, "jne"))
		{
		bJump = true;
		pTok->lType = JNE;
		}
	else if (!strcmp(pTok->psz, "jnz"))
		{
		bJump = true;
		pTok->lType = JNE;
		}
	else if (!strcmp(pTok->psz, "add"))
		pTok->lType = ADD;
	else if (!strcmp(pTok->psz, "sub"))
		pTok->lType = SUB;
	else if (!strcmp(pTok->psz, "mul"))
		pTok->lType = MUL;
	else if (!strcmp(pTok->psz, "div"))
		pTok->lType = DIV;
	else if (!strcmp(pTok->psz, "shl"))
		pTok->lType = SHL;
	else if (!strcmp(pTok->psz, "sar"))
		pTok->lType = SAR;
	/*else if (!strcmp(pTok->psz, "rol"))
	else if (!strcmp(pTok->psz, "ror"))*/
	else if (!strcmp(pTok->psz, "cmp"))
		pTok->lType = CMP;
	else if (!strcmp(pTok->psz, "sete"))
		pTok->lType = CMP;
	else if (!strcmp(pTok->psz, "not"))
		pTok->lType = NOT;
	else if (!strcmp(pTok->psz, "neg"))
		pTok->lType = NEG;
	else if (!strcmp(pTok->psz, "and"))
		pTok->lType = AND;
	else if (!strcmp(pTok->psz, "or"))
		pTok->lType = OR;
	else if (!strcmp(pTok->psz, "xor"))
		pTok->lType = XOR;
	else if (!strcmp(pTok->psz, "setne"))
		pTok->lType = SETNE;
	else if (!strcmp(pTok->psz, "setl"))
		pTok->lType = SETL;
	else if (!strcmp(pTok->psz, "setge"))
		pTok->lType = SETGE;
	else if (!strcmp(pTok->psz, "setg"))
		pTok->lType = SETG;
	else if (!strcmp(pTok->psz, "setle"))
		pTok->lType = SETLE;
	else if (!strcmp(pTok->psz, "land"))
		pTok->lType = LAND;
	else if (!strcmp(pTok->psz, "lor"))
		pTok->lType = LOR;
	else if (!strcmp(pTok->psz, "lxor"))
		pTok->lType = LXOR;
	else if (!strcmp(pTok->psz, "movenow"))
		pTok->lType = MOVENOW;
	else if (!strcmp(pTok->psz, "turnnow"))		
		pTok->lType = TURNNOW;
	else if (!strcmp(pTok->psz, "_cobalt"))
		pTok->lType = COBALT_OPCODE;
	else if (!strcmp(pTok->psz, "invoke"))
		{
		pTok->lType = INVOKE;
		bInvoke = true;
		}
	else if (/*!strcmp(pTok->psz, "string") || */!strcmp(pTok->psz, "pushstring"))
		{
		pTok->lType = PUSH_STRING;
		bPushString = true;
		}
	else if (!strcmp(pTok->psz, "string") || !strcmp(pTok->psz, "dstr"))
		{
		pTok->lType = OUTPUT_STRING;
		bPushString = true;
		bOutputStringWithoutPush = true;
		}
	else { pTok->lType = KW_UNRECOGNIZED; return NULL; }
	
	return pTok->pNext;
	}

_token *ArgList(_token *p)
	{
	_token *pTok = p;
	_token *pResult;
	bDecl = true;

Start:

	pResult = DeclArg(pTok);
	if (!pResult) { bDecl = false; return NULL; }

	pTok = pResult;
	if ((strlen(pTok->psz) == 1) && pTok->psz[0] == ',')
		{
		pTok->lType = COMMA;
		pTok = pTok->pNext;
		goto Start;
		}

	bDecl = false;
	return pResult;
	}

_token *VarList(_token *p)
	{
	_token *pTok = p;
	_token *pResult;
	bDecl = true;

Start:
	pResult = DeclVar(pTok);
	if (!pResult) { bDecl = false; return NULL; }
	pResult->pParent->lType = CREATELOCALVAR;

	pTok = pResult;
	if ((strlen(pTok->psz) == 1) && pTok->psz[0] == ',')
		{
		pTok->lType = COMMA;
		pTok = pTok->pNext;
		goto Start;
		}

	bDecl = false;
	return pResult;
	}

_token *Quoted(_token *p)
	{
	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	_token *pTok = p;
	_token *pResult;

	if (!((strlen(pTok->psz) == 1) && pTok->psz[0] == '\"')) return NULL;

	pTok->lType = QUOTEMARK;
	pTok = pTok->pNext;
	pResult = Arg(pTok);

	if (!pResult) return NULL;

	pTok = pResult;

	if (!((strlen(pTok->psz) == 1) && pTok->psz[0] == '\"')) return NULL;

	pTok->lType = QUOTEMARK;

	return pTok->pNext;
	}

_token *DeclArg(_token *p)
	{
	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	_token *pTok = p;
	_token *pResult;

	lDeclType = 3;

	pResult = Identifier(pTok);
	return pResult;
	}

_token *DeclVar(_token *p)
	{
	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	_token *pTok = p;
	_token *pResult;

	lDeclType = 3;

	pResult = Identifier(pTok);
	return pResult;
	}

long lPosDLLExportName;
long lPosStringName;

_token *Arg(_token *p)
	{
	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	_token *pTok = p;
	_token *pResult;

	if (bPushString)
		{
		lPosStringName = 0;
		bPushString = false;

		pResult = IsValidString(pTok);
		if (!pResult)
			{
			bError = true;
			lnErrors++;
			lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
			lpErrors[lnErrors - 1] = INVALID_STRING;
			lpErrors[lnErrors - 1] |= (pTok->lLine << 16);

			if (pPushStrString)
				{
				delete[] pPushStrString;
				pPushStrString = NULL;
				}

			return NULL;
			}
		
		bool bFoundString = false;

		for (long a = 0; (a < lnStringLinks) && (!bFoundString || bOutputStringWithoutPush); a++)
			{
			if (!strcmp(LinkedStrings[a].psz, pTok->psz))
				{
				bFoundString = true;
				for (long b = 0; b < a; b++)
					lPosStringName += strlen(LinkedStrings[a].psz) + 1;
				}
			}

		if (!bFoundString || bOutputStringWithoutPush)
			{
			lnStringLinks++;
			LinkedStrings = (_stringlink *)realloc(LinkedStrings, lnStringLinks * sizeof(_stringlink));
			LinkedStrings[lnStringLinks - 1].psz = new char [strlen(pTok->psz) + 1];
			strcpy(LinkedStrings[lnStringLinks - 1].psz, pPushStrString);

			for (long b = 0; b < (lnStringLinks - 1); b++)
				lPosStringName += strlen(LinkedStrings[b].psz) + 1;

			if (bFoundString)
				addwarning(ADDING_STRING_ALREADY_EXISTS, pTok);
			}

		if (bOutputStringWithoutPush)
			{
			pTok->lType = STRING_FOR_OUTPUT_ONLY;
			bOutputStringWithoutPush = false;
			}
		else
			pTok->lType = STRING_LINK;

		pTok->lMoreInfo = lPosStringName;

		return pTok->pNext;
		}

	if (bInvoke)
		{
		lPosDLLExportName = 0;
		bInvoke = false;

		pResult = IsValidInvokeIdentifier(pTok);
		if (!pResult)
			{
			bError = true;
			lnErrors++;
			lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
			lpErrors[lnErrors - 1] = INVALID_DLLLINK_SYMBOL;
			lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
			return NULL;
			}
		
		bool bFoundLink = false;

		for (long a = 0; (a < lnDLLLinks) && !bFoundLink; a++)
			{
			if (!strcmp(LinkedDLLProcs[a].psz, pTok->psz))
				{
				bFoundLink = true;
				for (long b = 0; b < a; b++)
					lPosDLLExportName += strlen(LinkedDLLProcs[a].psz) + 1;
				}
			}

		if (!bFoundLink)
			{
			lnDLLLinks++;
			LinkedDLLProcs = (_dlllink *)realloc(LinkedDLLProcs, lnDLLLinks * sizeof(_dlllink));
			LinkedDLLProcs[lnDLLLinks - 1].psz = new char [strlen(pTok->psz) + 1];
			strcpy(LinkedDLLProcs[lnDLLLinks - 1].psz, pTok->psz);

			for (long b = 0; b < (lnDLLLinks - 1); b++)
				lPosDLLExportName += strlen(LinkedDLLProcs[b].psz) + 1;
			}

		pTok->lType = DLL_LINK;
		pTok->lMoreInfo = lPosDLLExportName;

		return pTok->pNext;
		}

	if ((strlen(p->psz) == 1) && pTok->psz[0] == ';') return p;

	if (bJump)
		{
		pResult = LabelName(pTok);
		if (pResult) return pResult;
		}

	pResult = IsValidIdentifier(pTok);

	if (pResult)
		{
		bool bFound = false; //we must find the symbol in the symtab
		
		if (!strcmp(pTok->psz, "x-axis")){ bFound = true; pTok->lType = 0 | 16384; }
		else if (!strcmp(pTok->psz, "y-axis")){ bFound = true; pTok->lType = 1 | 16384; }
		else if (!strcmp(pTok->psz, "z-axis")){ bFound = true; pTok->lType = 2 | 16384; }

		_symtab *pSymtab;
		for (long i = 0; (i < 6) && !bFound; i++)
			{
			pSymtab = &symtab[i];
			while (pSymtab && pSymtab->tok.psz && !bFound)
				{
				if (!strcmp(pSymtab->tok.psz, pTok->psz))
					{
					bFound = true;
					pTok->lType = pSymtab->lIndex | 16384;
					}
				else
					pSymtab = pSymtab->pNext;
				}
			}
		if (!bFound)
			{
			lnErrors++;

			lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
			lpErrors[lnErrors - 1] = UNDEFINED_TERM;
			lpErrors[lnErrors - 1] |= (pTok->lLine << 16);

				{
				bool bFound = false;
				_symtab *ps = &symtab[6];

				for (long i = 0; (i < lSymbolIndex[6]) && !bFound; i++)
					{
					if (!strcmp(ps->tok.psz, pTok->psz)) bFound = true;
					else ps = ps->pNext;
					}

				if (!bFound)
					{
					ps->tok.psz = new char[strlen(pTok->psz) + 1];
					strcpy(ps->tok.psz, pTok->psz);
					ps->pNext = (_symtab *)malloc(sizeof(_symtab));
					ps->pNext->tok.psz = NULL;
					ps->pNext->pParent = ps;
					ps->pNext->pNext = NULL;
					ps->lIndex = lSymbolIndex[6]++;
					}

				lpErrors[lnErrors - 1] |= (ps->lIndex << 8);
				}

			return NULL; //didn't find it
			}
		}

	if (pResult) return pResult->pNext;

	pResult = Number(pTok);
	if (pResult) return pResult;	

	return pTok->pNext;
	}

_token *Number(_token *p)
	{
	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	_token *pTok = p;
	_token *pResult = Int(pTok);

	if (!pResult) pResult = Float(pTok);

	return pResult;
	}

_token *LabelName(_token *p)
	{
	_token *pTok = p;

	_symtab *ps = &symtab[5];
	bool bFound = false;

	for (long i = 0; i < lSymbolIndex[5]; i++)
		{
		if (!strcmp(ps->tok.psz, pTok->psz)) bFound = true;
		ps = ps->pNext;
		}

	if (!bFound && bJump)
		{
		ps->pNext = (_symtab *)malloc(sizeof(_symtab));
		ps->pNext->tok.psz = NULL;
		ps->pNext->pParent = ps;
		ps->pNext->pNext = NULL;
		ps->lIndex = lSymbolIndex[5]++;
		ps->tok.psz = (char *)malloc(strlen(pTok->psz) + 1);
		strcpy(ps->tok.psz, pTok->psz);
		ps = ps->pNext;
		bJump = false;
		}

	pTok->lType = LBL_JMP;
	return pTok->pNext;
	}

_token *Int(_token *p)
	{
	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	bool bHex = false;

	long lType = 0;
	long lConversionTimes = 0;

	_token *pTok = p;

	long lLen = strlen(pTok->psz);

	for (long i = 0; i < lLen; i++)
		{
		if (!isdigit(pTok->psz[i]) && (pTok->psz[i] != '-'))
			{
			if ((pTok->psz[i] != 'A') && (pTok->psz[i] != 'B') && (pTok->psz[i] != 'C') && (pTok->psz[i] != 'D') && (pTok->psz[i] != 'E') && (pTok->psz[i] != 'F') &&
				(pTok->psz[i] != 'a') && (pTok->psz[i] != 'b') && (pTok->psz[i] != 'c') && (pTok->psz[i] != 'd') && (pTok->psz[i] != 'e') && (pTok->psz[i] != 'f'))
				{
				if (lConversionTimes > 2)
					{
					bError = true;
					lnErrors++;
					lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
					lpErrors[lnErrors - 1] = TOO_MANY_CONVERSIONS;
					lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
					return NULL;
					}

				if (pTok->psz[i] == 'h')
					{
					/*pTok->lType = HEX_INT;
					pTok->psz[i] = NULL;
					return pTok->pNext;*/
					if (lType & 1)
						{
						bError = true;
						lnErrors++;
						lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
						lpErrors[lnErrors - 1] = CANT_CONVERT_TWICE;
						lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
						return NULL;
						}

					lType |= 1;
					}
				else if (pTok->psz[i] == 'n')
					{
					/*pTok->lType = ANGULAR_INT;
					pTok->psz[i] = NULL;
					return pTok->pNext;*/
					if (lType & 2)
						{
						bError = true;
						lnErrors++;
						lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
						lpErrors[lnErrors - 1] = CANT_CONVERT_TWICE;
						lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
						}

					if (lType & 4)
						{
						bError = true;
						lnErrors++;
						lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
						lpErrors[lnErrors - 1] = INCOMPATIBLE_CONVERSION_TYPES;
						lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
						}

					if (bError) return NULL;

					lType |= 2;
					}
				else if (pTok->psz[i] == 'l')
					{
					/*pTok->lType = LINEAR_INT;
					pTok->psz[i] = NULL;
					return pTok->pNext;*/
					if (lType & 4)
						{
						bError = true;
						lnErrors++;
						lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
						lpErrors[lnErrors - 1] = CANT_CONVERT_TWICE;
						lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
						}

					if (lType & 2)
						{
						bError = true;
						lnErrors++;
						lpErrors = (long *)realloc(lpErrors, lnErrors * 4);
						lpErrors[lnErrors - 1] = INCOMPATIBLE_CONVERSION_TYPES;
						lpErrors[lnErrors - 1] |= (pTok->lLine << 16);
						}

					if (bError) return NULL;

					lType |= 4;
					}
				else
					return NULL;

				lConversionTimes++;
				}				
			else { bHex = true; }
			}
		}

	if (bHex && ((lType & 1) == 0)) //if we got out of the loop without finding an 'h'
		{
		return NULL;
		}

	pTok->lType = INT;

	if (lType)
		{
		if (lType & 1)
			{
			if (lType & 2) pTok->lType = ANGULAR_HEX_INT;
			else if (lType & 4)	pTok->lType = LINEAR_HEX_INT;
			else pTok->lType = HEX_INT;
			}
		else if (lType & 2) pTok->lType = ANGULAR_INT;
		else if (lType & 4)	pTok->lType = LINEAR_INT;
		}

	return pTok->pNext;
	}

_token *Float(_token *p)
	{
	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	_token *pTok = p;
	_token *pResult;
	bool bDec = false;
	long l = strlen(pTok->psz);

	for (long i = 0; i < l; i++)
		{
		if (!isdigit(pTok->psz[i]) && (pTok->psz[i] != '-'))
			{
			if (pTok->psz[i] == '.')
				{
				if (bDec == false) bDec = true;
				else return NULL;
				}
			else if (pTok->psz[i] == 'n')
					{
					pTok->lType = ANGULAR_FLOAT;
					pTok->psz[i] = NULL;
					return pTok->pNext;
					}
			else if (pTok->psz[i] == 'l')
					{
					pTok->lType = LINEAR_FLOAT;
					pTok->psz[i] = NULL;
					return pTok->pNext;
					}
			else return NULL;
			}
		}

	if (!bDec) return NULL;

	pTok->lType = FLOAT;
	pResult = pTok->pNext;

	return pResult;
	}

_token *Label(_token *p)
	{
	if (!p->psz)
		{
		bEOF = true;
		return NULL;
		}

	lDeclType = 5;
	_token *pTok = p;
	_token *pResult = IsValidIdentifier(pTok);

	if (!pResult) return NULL;
	pResult = pResult->pNext;

	if ((strlen(pResult->psz) == 1) && pResult->psz[0] == ':')
		{
		pResult->pParent->lType = LABEL;
		pResult->lType = NOCODE;
		bool bFound = false;
		_symtab *ps = &symtab[5];
		for (long i = 0; i < lSymbolIndex[5]; i++)
			{
			if (!strcmp(ps->tok.psz, pResult->pParent->psz)) bFound = true;
			ps = ps->pNext;
			}

		if (!bFound)
			{
			ps->pNext = (_symtab *)malloc(sizeof(_symtab));
			ps->pNext->tok.psz = NULL;
			ps->pNext->pParent = ps;
			ps->pNext->pNext = NULL;
			ps->lIndex = lSymbolIndex[5]++;
			ps->tok.psz = (char *)malloc(strlen(pTok->psz) + 1);
			strcpy(ps->tok.psz, pTok->psz);
			ps = ps->pNext;
			}
		pResult = pResult->pNext;
		}
	else pResult = NULL;

	return pResult;
	}